#include <ShaderManager.h>

#include <Mouse.h>

#include <Keyboard.h>

#include "Config.h"

using namespace Render;

using namespace Math;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ()   ()

Mouse mouse ( 0.01F );

Keyboard keyboard ( 0.05F );

Camera * camera = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

void MouseMove ( int x, int y )
{
	mouse.MouseMove ( x, y );
}
		
void MouseButton ( int button, int state )
{
	mouse.StateChange ( button, state );
}

void KeyButton ( int key, int state )
{
	keyboard.StateChange ( key, state );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int main ( void )
{
	glfwInit ( );

	//---------------------------------------------------------------------------------------------

	cout << "Do you want to run program in fullscreen mode? [Y/N]" << endl;

	int choice = getchar ( );

	int width = 750, height = 750, mode = GLFW_WINDOW;

	//---------------------------------------------------------------------------------------------

	if ( choice == 'Y' || choice == 'y' )
	{
		GLFWvidmode vidmode;

		glfwGetDesktopMode ( &vidmode );

		width = vidmode.Width;
		
		height = vidmode.Height;

		mode = GLFW_FULLSCREEN;
	}

	//---------------------------------------------------------------------------------------------

    if( !glfwOpenWindow ( width, height, 0, 0, 0, 0, 0, 0, mode ) )
    {
        glfwTerminate ( ); exit ( 0 );
	}

	//---------------------------------------------------------------------------------------------

    glfwSwapInterval ( 0 );

	glfwSetMousePosCallback ( MouseMove );

	glfwSetMouseButtonCallback ( MouseButton );

	glfwSetKeyCallback ( KeyButton );

	//---------------------------------------------------------------------------------------------
	
	glMatrixMode ( GL_PROJECTION );
	
	glLoadIdentity ( );

#ifdef RENDER_TRUE_STEREO
	
	glOrtho ( -2.0F, 2.0F, -1.0F, 1.0F, -1.0F, 1.0F );
	
#else

	glOrtho ( -1.0F, 1.0F, -1.0F, 1.0F, -1.0F, 1.0F );

#endif

	glMatrixMode ( GL_MODELVIEW );
	
	glLoadIdentity ( );	

	//---------------------------------------------------------------------------------------------

	camera = new Camera ( Vector3D ( 0.0F, 0.0F, -18.0F ),
		                  Vector3D ( 0.0F, 0.0F, 0.0F ) );

	camera->SetFrustum ( );

	//---------------------------------------------------------------------------------------------

	ShaderManager * manager = new ShaderManager ( );

	manager->LoadVertexShader ( "Render.vs" );

	const char ** files = new const char * [2];
	
	files [0] = "Config.h"; files [1] = "Render.fs";
	
	manager->LoadFragmentShader ( files, 2 );

	manager->BuildProgram ( );

	//---------------------------------------------------------------------------------------------
	
	bool running = GL_TRUE;

	int frames = 0;

	char caption [100];

	double start = glfwGetTime ( );

	//---------------------------------------------------------------------------------------------

    while ( running )
    {
        double time = glfwGetTime ( );

        if ( ( time - start ) > 1.0 || frames == 0 )
        {
            double fps = frames / ( time - start );

            sprintf_s ( caption, "Implicity Surface Visualization - %.1f FPS", fps );

            glfwSetWindowTitle ( caption );

            start = time;

            frames = 0;
        }

        frames++;

		//-----------------------------------------------------------------------------------------

        glfwGetWindowSize ( &width, &height );

        height = max ( height, 1 );

#ifdef RENDER_TRUE_STEREO

		camera->SetViewport ( width / 2, height );

#else

		camera->SetViewport ( width, height );

#endif

		glViewport ( 0, 0, width, height );
		
		//-----------------------------------------------------------------------------------------

		mouse.Apply ( camera );

		keyboard.Apply ( camera );

		//-----------------------------------------------------------------------------------------

		glClear ( GL_COLOR_BUFFER_BIT );

		manager->Bind ( );

		manager->SetUniformFloat ( "Time", time );

		camera->SetShaderData ( manager );

#ifdef RENDER_TRUE_STEREO

		manager->SetUniformFloat ( "Time", time );
	
		glBegin ( GL_QUADS );

			glVertex2f ( -2.0F, -1.0F );
			glVertex2f ( -2.0F,  1.0F );
			glVertex2f (  0.0F,  1.0F );
			glVertex2f (  0.0F, -1.0F );

		glEnd ( );

		manager->SetUniformFloat ( "Time", time );
	
		glBegin ( GL_QUADS );

			glVertex2f ( 0.0F, -1.0F );
			glVertex2f ( 0.0F,  1.0F );
			glVertex2f ( 2.0F,  1.0F );
			glVertex2f ( 2.0F, -1.0F );

		glEnd ( );
	
#else

		glBegin ( GL_QUADS );

			glVertex2f ( -1.0F, -1.0F );
			glVertex2f ( -1.0F,  1.0F );
			glVertex2f (  1.0F,  1.0F );
			glVertex2f (  1.0F, -1.0F );

		glEnd ( );

#endif

		manager->Unbind ( );

        glfwSwapBuffers ( );

		//-----------------------------------------------------------------------------------------
		
		running = !glfwGetKey ( GLFW_KEY_ESC ) && glfwGetWindowParam ( GLFW_OPENED );
	}

    glfwTerminate ( ); exit ( 0 );
}