#include <ShaderManager.h>

#include <Mouse.h>

#include <Keyboard.h>

#include "Config.h"

using namespace Render;

using namespace Math;

///////////////////////////////////////////////////////////////////////////////////////////////////
//     4D  

struct Fractal
{
	public:

		//    
		Vector4D Center;      
		
		//         
		static const float EscapeThreshold;
		
		//      
		static const float Epsilon ;
		
		//         
		static const float Delta;
		
		//       / 
		static const int Maximum;

		Fractal ( Vector4D center = Vector4D ( -0.2F, 0.2F, 0.5F, 0.5F ) )
		{
			Center = center;
		}

        void SetShaderData ( ShaderManager * manager )
        {
        	manager->SetUniformFloat ( "JuliaFractal.EscapeThreshold", EscapeThreshold );
        	
        	manager->SetUniformFloat ( "JuliaFractal.Epsilon", Epsilon );
        	
        	manager->SetUniformFloat ( "JuliaFractal.Delta", Delta );
        	
        	manager->SetUniformInteger ( "JuliaFractal.Maximum", Maximum );
        	
        	manager->SetUniformVector ( "JuliaFractal.Center", Center );
		}
};

const float Fractal :: EscapeThreshold = 10.0F;

const float Fractal :: Epsilon = 0.005F;

const float Fractal :: Delta = 0.001F;

const int Fractal :: Maximum = 100;

///////////////////////////////////////////////////////////////////////////////////////////////////
//      

Vector4D centers [] = { Vector4D ( -1.2F, 0.2F, 0.3F, 0.0F ),
                        Vector4D ( -0.2F, 0.5F, 0.3F, 0.0F ),
                        Vector4D ( -0.1F, 0.8F, 0.6F, 0.0F ),
						Vector4D ( 0.1F, 0.8F, 0.6F, 0.0F ),
						Vector4D ( -1.1F, 0.2F, 0.1F, 0.0F ),
						Vector4D ( -0.4F, 0.1F, 0.8F, 0.1 ) };

int curr = 0;

int next = 1;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ()   ()

Mouse mouse ( 0.01F );

Keyboard keyboard ( 0.05F );

Camera * camera = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//       

Fractal * fractal = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

void MouseMove ( int x, int y )
{
	mouse.MouseMove ( x, y );
}
		
void MouseButton ( int button, int state )
{
	mouse.StateChange ( button, state );
}

void KeyButton ( int key, int state )
{
	keyboard.StateChange ( key, state );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int main ( void )
{
	glfwInit ( );

	//---------------------------------------------------------------------------------------------

	cout << "Do you want to run program in fullscreen mode? [Y/N]" << endl;

	int choice = getchar ( );

	int width = 750, height = 750, mode = GLFW_WINDOW;

	//---------------------------------------------------------------------------------------------

	if ( choice == 'Y' || choice == 'y' )
	{
		GLFWvidmode vidmode;

		glfwGetDesktopMode ( &vidmode );

		width = vidmode.Width;
		
		height = vidmode.Height;

		mode = GLFW_FULLSCREEN;
	}

	//---------------------------------------------------------------------------------------------

    if( !glfwOpenWindow ( width, height, 0, 0, 0, 0, 0, 0, mode ) )
    {
        glfwTerminate ( ); exit ( 0 );
	}

	//---------------------------------------------------------------------------------------------

    glfwSwapInterval ( 0 );

	glfwSetMousePosCallback ( MouseMove );

	glfwSetMouseButtonCallback ( MouseButton );

	glfwSetKeyCallback ( KeyButton );

	//---------------------------------------------------------------------------------------------
	
	glMatrixMode ( GL_PROJECTION );
	
	glLoadIdentity ( );

#ifdef RENDER_TRUE_STEREO
	
	glOrtho ( -2.0F, 2.0F, -1.0F, 1.0F, -1.0F, 1.0F );
	
#else

	glOrtho ( -1.0F, 1.0F, -1.0F, 1.0F, -1.0F, 1.0F );

#endif

	glMatrixMode ( GL_MODELVIEW );
	
	glLoadIdentity ( );	

	//---------------------------------------------------------------------------------------------

	camera = new Camera ( Vector3D ( 0.0F, 0.0F, -18.0F ),
		                  Vector3D ( 0.0F, 0.0F, 0.0F ) );

	camera->SetFrustum ( );

	//---------------------------------------------------------------------------------------------


	fractal = new Fractal ( );

	//---------------------------------------------------------------------------------------------

	ShaderManager * manager = new ShaderManager ( );

	manager->LoadVertexShader ( "Julia.vs" );

	const char ** files = new const char * [2];
	
	files [0] = "Config.h"; files [1] = "Julia.fs";
	
	manager->LoadFragmentShader ( files, 2 );

	manager->BuildProgram ( );

	//---------------------------------------------------------------------------------------------
	
	bool running = GL_TRUE;

	int frames = 0;

	char caption [100];

	double start = glfwGetTime ( );

	//---------------------------------------------------------------------------------------------

    while ( running )
    {
        double time = glfwGetTime ( );

        if ( ( time - start ) > 4.0 || frames == 0 )
        {
            double fps = frames / ( time - start );

            sprintf_s ( caption, "Julia Fractal Visualization [%.1f %.1f %.1f %.1f] - %.1f FPS",
				        fractal->Center.X,
						fractal->Center.Y,
						fractal->Center.Z,
						fractal->Center.W,
						fps );

            glfwSetWindowTitle ( caption );

            start = time;

            frames = 0;
			
			curr = next;

			next = ( curr + 1 ) % 6;
        }

        frames++;

		//-----------------------------------------------------------------------------------------

		fractal->Center = Mix ( centers [curr],
			                    centers [next],
								Fract ( 0.25F * ( time - start ) ) );

		//-----------------------------------------------------------------------------------------

        glfwGetWindowSize ( &width, &height );

        height = max ( height, 1 );

#ifdef RENDER_TRUE_STEREO

		camera->SetViewport ( width / 2, height );

#else

		camera->SetViewport ( width, height );

#endif

		glViewport ( 0, 0, width, height );
		
		//-----------------------------------------------------------------------------------------

		mouse.Apply ( camera );

		keyboard.Apply ( camera );

		//-----------------------------------------------------------------------------------------

		glClear ( GL_COLOR_BUFFER_BIT );

		manager->Bind ( );

		camera->SetShaderData ( manager );

		fractal->SetShaderData ( manager );

#ifdef RENDER_TRUE_STEREO

		glBegin ( GL_QUADS );

			glVertex2f ( -2.0F, -1.0F );
			glVertex2f ( -2.0F,  1.0F );
			glVertex2f (  0.0F,  1.0F );
			glVertex2f (  0.0F, -1.0F );
			
			glVertex2f ( 0.0F, -1.0F );
			glVertex2f ( 0.0F,  1.0F );
			glVertex2f ( 2.0F,  1.0F );
			glVertex2f ( 2.0F, -1.0F );

		glEnd ( );
	
#else

		glBegin ( GL_QUADS );

			glVertex2f ( -1.0F, -1.0F );
			glVertex2f ( -1.0F,  1.0F );
			glVertex2f (  1.0F,  1.0F );
			glVertex2f (  1.0F, -1.0F );

		glEnd ( );

#endif

		manager->Unbind ( );

        glfwSwapBuffers ( );

		//-----------------------------------------------------------------------------------------
		
		running = !glfwGetKey ( GLFW_KEY_ESC ) && glfwGetWindowParam ( GLFW_OPENED );
	}

    glfwTerminate ( ); exit ( 0 );
}