#include <ShaderManager.h>

#include <Texture2D.h>

#include <Vector3D.h>

#include <FrameBuffer.h>

//-------------------------------------------------------------------------------------------------

using namespace Math;

using namespace Render;

//-------------------------------------------------------------------------------------------------

const int N = 5;

const int M = 128;

//-------------------------------------------------------------------------------------------------

TextureData2D * dataA = NULL;   //      A

TextureData2D * dataB = NULL;   //      B

TextureData2D * dataC = NULL;   //      C

//-------------------------------------------------------------------------------------------------

Texture2D * textureA = NULL;   //     A

Texture2D * textureB = NULL;   //     B      

Texture2D * textureC = NULL;   //     C

//-------------------------------------------------------------------------------------------------

ShaderManager * additionProgram = NULL;   //    

//-------------------------------------------------------------------------------------------------

FrameBuffer * additionFrameBuffer = NULL;   //      

//-------------------------------------------------------------------------------------------------

float RangedRand ( int min, int max )
{
   return ( float ) rand ( ) / ( RAND_MAX + 1 ) * ( max - min ) + min;
}

//-------------------------------------------------------------------------------------------------

void SetupMatrices ( )
{
	{
		dataA = new TextureData2D ( N, M, 1 );

		dataB = new TextureData2D ( N, M, 1 );     		
                
		dataC = new TextureData2D ( N, M, 1 );
	}
	
	for ( int i = 0; i < N; i++ )
	{
		for ( int j = 0; j < M; j++ )
		{
			dataA->Pixel < float > ( i, j ) = RangedRand ( -100.0F, 100.0F );

			dataB->Pixel < float > ( i, j ) = RangedRand ( -100.0F, 100.0F );
		}
	}
}

//-------------------------------------------------------------------------------------------------

void SetupKernels ( )
{
	{
		textureA = new Texture2D ( dataA, 0, GL_TEXTURE_RECTANGLE_ARB );
				
		textureB = new Texture2D ( dataB, 1, GL_TEXTURE_RECTANGLE_ARB );

		textureC = new Texture2D ( dataC, 2, GL_TEXTURE_RECTANGLE_ARB );
					
		textureA->Setup ( );
				
		textureB->Setup ( );
				
		textureC->Setup ( );			
	}
	
	{
		bool result = true;
		
		additionProgram = new ShaderManager ( );
		
		if ( !additionProgram->LoadVertexShader ( "Addition.vs" ) )
			result = false;
		
		if ( !additionProgram->LoadFragmentShader ( "Addition.fs" ) )
			result = false;
		
		if ( !additionProgram->BuildProgram ( ) )
			result = false;
		
		if ( !result )
			cout << "Compute program error!" << endl;
		
		additionProgram->Bind ( );
		
		additionProgram->SetTexture ( "TextureA", textureA );
		
		additionProgram->SetTexture ( "TextureB", textureB );
		
		additionProgram->Unbind ( );
	}
	
	{
		additionFrameBuffer = new FrameBuffer ( );
		
		additionFrameBuffer->ColorBuffers.push_back ( textureC );

		additionFrameBuffer->Setup ( );
	}
}

//-------------------------------------------------------------------------------------------------

void GenerateStream ( )
{
	{
		glViewport ( 0, 0, N, M );    
				
		glMatrixMode ( GL_PROJECTION );
	            
		glLoadIdentity ( ); 
		
		glOrtho ( 0.0F, 1.0F, 0.0F, 1.0F, 0.0F, 1.0F );
		
		glMatrixMode ( GL_MODELVIEW );
		
		glLoadIdentity ( );				
	}

	{
		glClear ( GL_COLOR_BUFFER_BIT );
		
		glBegin ( GL_QUADS );
		{
			glTexCoord2f ( 0.0F, 0.0F );   glVertex2f ( 0.0F, 0.0F );
			glTexCoord2f ( 0.0F,    M );   glVertex2f ( 0.0F, 1.0F );
			glTexCoord2f ( N,       M );   glVertex2f ( 1.0F, 1.0F );
			glTexCoord2f ( N,    0.0F );   glVertex2f ( 1.0F, 0.0F );
		}
		glEnd ( ); 
	}
}

//-------------------------------------------------------------------------------------------------

void StartKernels ( )
{
	additionFrameBuffer->Bind ( );

	additionProgram->Bind ( );
	
	GenerateStream ( );
	
	additionFrameBuffer->FetchOutcome ( );
}

//-------------------------------------------------------------------------------------------------

void OutputResult ( )
{
	for ( int i = 0; i < N; i++ )
	{
		for ( int j = 0; j < M; j++ )
		{
			cout << "A[" << i << "][" << j << "] = " << dataA->Pixel < float > ( i, j ) << endl;

			cout << "B[" << i << "][" << j << "] = " << dataB->Pixel < float > ( i, j ) << endl;

			cout << "C[" << i << "][" << j << "] = " << dataC->Pixel < float > ( i, j ) << endl;
		}
	}
}

//-------------------------------------------------------------------------------------------------

int main ( void )
{	
	char title [200];

	//---------------------------------------------------------------------------------------------

    glfwInit ( );

	if( !glfwOpenWindow ( 0, 0, 0, 0, 0, 0, 0, 0, GLFW_WINDOW ) )
    {
        glfwTerminate ( );

        return 0;
	}

    glfwSwapInterval ( 0 );

	//---------------------------------------------------------------------------------------------
	
	SetupMatrices ( );
	
	SetupKernels ( );

	//---------------------------------------------------------------------------------------------
	
	bool running = true;

	int frames = 0;

	float start = glfwGetTime ( );

	//---------------------------------------------------------------------------------------------

    while ( running )
	{
        float curent = glfwGetTime ( );

        if ( ( curent - start ) > 1.0F || frames == 0 )
        {
            float fps = ( float ) frames / ( curent - start );

            sprintf_s ( title, "Matrix Addition - %.1f iterations / sec", fps );

            glfwSetWindowTitle ( title );

            start = curent;

            frames = 0;
        }

        frames++;
		
		//-----------------------------------------------------------------------------------------

		StartKernels ( );
		
		OutputResult ( );

		//-----------------------------------------------------------------------------------------

        glfwSwapBuffers ( );		
		
		running = false;//!glfwGetKey ( GLFW_KEY_ESC ) && glfwGetWindowParam ( GLFW_OPENED );
	}

    glfwTerminate ( );

    return 0;
}