#include <Mouse.h>

#include <Keyboard.h>

#include <FrameBuffer.h>

using namespace Math;

using namespace Render;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ()   ()

Mouse mouse ( 0.01F );

Keyboard keyboard ( 0.05F );

Camera * camera = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int countX = 128;

int countY = 128;

///////////////////////////////////////////////////////////////////////////////////////////////////
//   ()   ,     

TextureData2D * positions = NULL;

TextureData2D * velocities = NULL;

TextureData2D * colors = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

Vector4D positionAverage ( 0.0F, 0.0F, 0.0F, 20.0F );

Vector4D positionDispersion ( 2.0F, 2.0F, 2.0F, 10.0F );

Vector3D velocityAverage ( 5.0F, 5.0F, 5.0F );

Vector3D velocityDispersion ( 0.0F, 0.0F, 0.0F );

Vector3D colorAverage ( 0.75F, 0.75F, 0.75F );

Vector3D colorDispersion ( 0.25F, 0.25F, 0.25F );

///////////////////////////////////////////////////////////////////////////////////////////////////
//       

Texture2D * currentPositionTexture = NULL;

Texture2D * currentVelocityTexture = NULL;      

Texture2D * nextPositionTexture = NULL;

Texture2D * nextVelocityTexture = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ( )

ShaderManager * firstProgram = NULL;

ShaderManager * secondProgram = NULL;

ShaderManager * renderProgram = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//      

FrameBuffer * firstFrameBuffer = NULL;

FrameBuffer * secondFrameBuffer = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//      [-1, +1]

float RandDouble ( void )
{
	return ( rand ( ) / ( ( float ) ( RAND_MAX ) + 1.0F ) ) * 2.0F - 1.0F;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//      

#define INIT_RANDOM_

#define INIT_SHELL

void SetupParticles ( void )
{
	positions = new TextureData2D ( countX, countY, 4 );

	velocities = new TextureData2D ( countX, countY, 3 );	

	colors = new TextureData2D ( countX, countY, 3 );

#if defined ( INIT_RANDOM )

	for ( int i = 0; i < countX; i++ )
	{
		for ( int j = 0; j < countY; j++ )
		{
			positions->Pixel<Vector4D> ( i, j ) =
				Vector4D ( positionAverage.X + positionDispersion.X * RandDouble ( ),
				           positionAverage.Y + positionDispersion.Y * RandDouble ( ),
						   positionAverage.Z + positionDispersion.Z * RandDouble ( ),
						   positionAverage.W + positionDispersion.W * RandDouble ( ) );

			velocities->Pixel<Vector3D> ( i, j ) =
				Vector3D ( velocityAverage.X + velocityDispersion.X * RandDouble ( ),
				           velocityAverage.Y + velocityDispersion.Y * RandDouble ( ),
						   velocityAverage.Z + velocityDispersion.Z * RandDouble ( ) );

			colors->Pixel<Vector3D> ( i, j ) =
				Vector3D ( colorAverage.X + colorDispersion.X * RandDouble ( ),
				           colorAverage.Y + colorDispersion.Y * RandDouble ( ),
						   colorAverage.Z + colorDispersion.Z * RandDouble ( ) );
		}
	}

#elif defined ( INIT_SHELL )

	float size = 2.0F;
	
	float scale = 12.0F * size;
	
	float inner = 2.5F * size;
	
	float outer = 4.0F * size;

	for ( int i = 0; i < countX; i++ )
	{
		for ( int j = 0; j < countY; j++ )
		{
			Vector3D point = Vector3D ( RandDouble ( ), RandDouble ( ), RandDouble ( ) );
			
			float length = Length ( point );

			while ( length > 1.0F )
			{
				point = Vector3D ( RandDouble ( ), RandDouble ( ), RandDouble ( ) );
				
				length = Length ( point );
			}
			
			point /= length;

			positions->Pixel<Vector4D> ( i, j ) = Vector4D (
				point.X * ( inner + ( outer - inner ) * rand ( ) / ( float ) RAND_MAX ),
				point.Y * ( inner + ( outer - inner ) * rand ( ) / ( float ) RAND_MAX ),
				point.Z * ( inner + ( outer - inner ) * rand ( ) / ( float ) RAND_MAX ),
				positionAverage.W );

			Vector3D axis = Vector3D :: AxisZ;
				
			if ( 1.0 - Dot ( point, axis ) < 1e-6 )
			{
				axis.X = point.Y;
				axis.Y = point.X;
				Normalize ( axis );
			}
			
			velocities->Pixel<Vector3D> ( i, j ) = scale *
				Cross ( positions->Pixel<Vector3D> ( i, j ), axis );

			colors->Pixel<Vector3D> ( i, j ) =
				Vector3D ( colorAverage.X + colorDispersion.X * RandDouble ( ),
				           colorAverage.Y + colorDispersion.Y * RandDouble ( ),
						   colorAverage.Z + colorDispersion.Z * RandDouble ( ) );
		}
	}

#endif
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//     ( )

void SetupKernels ( void )
{
	{
		currentPositionTexture = new Texture2D ( positions, 0, GL_TEXTURE_RECTANGLE_ARB );				
		currentVelocityTexture = new Texture2D ( velocities, 1, GL_TEXTURE_RECTANGLE_ARB );
		nextPositionTexture = new Texture2D ( positions, 2, GL_TEXTURE_RECTANGLE_ARB );				
		nextVelocityTexture = new Texture2D ( velocities, 3, GL_TEXTURE_RECTANGLE_ARB );

		currentPositionTexture->Setup ( );				
		currentVelocityTexture->Setup ( );				
		nextPositionTexture->Setup ( );		
		nextVelocityTexture->Setup ( );					
	}
	
	{
		firstProgram = new ShaderManager ( );
		
		if ( !firstProgram->LoadVertexShader ( "Compute.vs" ) )
		{
			cout << "ERROR! Failed to load compute vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !firstProgram->LoadFragmentShader ( "Compute.fs" ) )
		{
			cout << "ERROR! Failed to load compute fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !firstProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build compute shader program!" << endl; exit ( -1 );
		}	
		
		firstProgram->Bind ( );		
		firstProgram->SetTexture ( "CurrentPositionTexture", currentPositionTexture );		
		firstProgram->SetTexture ( "CurrentVelocityTexture", currentVelocityTexture );		
		firstProgram->Unbind ( );
	}
	
	{
		secondProgram = new ShaderManager ( );

		if ( !secondProgram->LoadVertexShader ( "Compute.vs" ) )
		{
			cout << "ERROR! Failed to load compute vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !secondProgram->LoadFragmentShader ( "Compute.fs" ) )
		{
			cout << "ERROR! Failed to load compute fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !secondProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build compute shader program!" << endl; exit ( -1 );
		}
		        
		secondProgram->Bind ( );		
		secondProgram->SetTexture ( "CurrentPositionTexture", nextPositionTexture );		
		secondProgram->SetTexture ( "CurrentVelocityTexture", nextVelocityTexture );		
		secondProgram->Unbind ( );
	}
	
	{
		renderProgram = new ShaderManager ( );

		if ( !renderProgram->LoadVertexShader ( "Render.vs" ) )
		{
			cout << "ERROR! Failed to load render vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !renderProgram->LoadFragmentShader ( "Render.fs" ) )
		{
			cout << "ERROR! Failed to load render fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !renderProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build render shader program!" << endl; exit ( -1 );
		}
		        
		renderProgram->Bind ( );		
		renderProgram->SetTexture ( "CurrentPositionTexture", currentPositionTexture );		
		renderProgram->Unbind ( );
	}
	
	{
		firstFrameBuffer = new FrameBuffer ( );		
		firstFrameBuffer->ColorBuffers.push_back ( nextPositionTexture );
		firstFrameBuffer->ColorBuffers.push_back ( nextVelocityTexture );
		firstFrameBuffer->Setup ( );
	}
	
	{
		secondFrameBuffer = new FrameBuffer ( );		
		secondFrameBuffer->ColorBuffers.push_back ( currentPositionTexture );		
		secondFrameBuffer->ColorBuffers.push_back ( currentVelocityTexture );
		secondFrameBuffer->Setup ( );
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//       ( = )

void GenerateStream ( int width, int height )
{
	{
		glViewport ( 0, 0, width, height );    
				
		glMatrixMode ( GL_PROJECTION );
	            
		glLoadIdentity ( ); 
		
		glOrtho ( 0.0F, 1.0F, 0.0F, 1.0F, -1.0F, 1.0F );
		
		glMatrixMode ( GL_MODELVIEW );
		
		glLoadIdentity ( );				
	}

	{
		glClear ( GL_COLOR_BUFFER_BIT );
		
		glBegin ( GL_QUADS );
		{
			glTexCoord2f (   0.0F,   0.0F );  glVertex2f ( 0.0F, 0.0F );
			glTexCoord2f (   0.0F, countY );  glVertex2f ( 0.0F, 1.0F );
			glTexCoord2f ( countX, countY );  glVertex2f ( 1.0F, 1.0F );
			glTexCoord2f ( countX,   0.0F );  glVertex2f ( 1.0F, 0.0F );
		}
		glEnd ( ); 
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

bool step = true;

void StartKernels ( )
{
	if ( step = !step )
	{
		firstFrameBuffer->Bind ( );

		firstProgram->Bind ( );
		
		GenerateStream ( countX, countY );

		firstFrameBuffer->Unbind ( );
	}
	else
	{
		secondFrameBuffer->Bind ( );

		secondProgram->Bind ( );

		GenerateStream ( countX, countY );

		secondFrameBuffer->Unbind ( );
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//       

void BuildList ( void )
{
	glNewList ( 1, GL_COMPILE );

	glBegin ( GL_QUADS );

	for ( int i = 0; i < countX; i++ )
	{
		for ( int j = 0; j < countY; j++ )
		{			
			glColor3fv ( colors->Pixel <Vector3D> ( i, j ) );
			
			glTexCoord2f ( i, j );

			glVertex2f ( -1.0F, -1.0F );								
			glVertex2f ( -1.0F,  1.0F );								
			glVertex2f (  1.0F,  1.0F );								
			glVertex2f (  1.0F, -1.0F );
		}
	}

	glEnd ( );

	glEndList ( );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  

void Draw ( int width, int height )
{
	mouse.Apply ( camera );

	keyboard.Apply ( camera );

	camera->SetViewport ( width, height );
            	
	camera->Setup ( );

	renderProgram->Bind ( );

	renderProgram->SetUniformVector ( "UpDirection", camera->GetUp ( ) );

	renderProgram->SetUniformVector ( "RightDirection", camera->GetSide ( ) );

	glClear ( GL_COLOR_BUFFER_BIT );

	glCallList ( 1 );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

void MouseMove ( int x, int y )
{
	mouse.MouseMove ( x, y );
}

void MouseButton ( int button, int state )
{
	mouse.StateChange ( button, state );
}

void KeyButton ( int key, int state )
{
	keyboard.StateChange ( key, state );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int main ( void )
{
	glfwInit ( );

	//---------------------------------------------------------------------------------------------

	cout << "Do you want to run program in fullscreen mode? [Y/N]" << endl;

	int choice = getchar ( );

	int width = 750, height = 750, mode = GLFW_WINDOW;

	//---------------------------------------------------------------------------------------------

	if ( choice == 'Y' || choice == 'y' )
	{
		GLFWvidmode vidmode;

		glfwGetDesktopMode ( &vidmode );

		width = vidmode.Width;
		
		height = vidmode.Height;

		mode = GLFW_FULLSCREEN;
	}

	//---------------------------------------------------------------------------------------------

    if( !glfwOpenWindow ( width, height, 0, 0, 0, 0, 0, 0, mode ) )
    {
        glfwTerminate ( ); exit ( 0 );
	}

	//---------------------------------------------------------------------------------------------

    glfwSwapInterval ( 0 );

	glfwSetMousePosCallback ( MouseMove );

	glfwSetMouseButtonCallback ( MouseButton );

	glfwSetKeyCallback ( KeyButton );

	//---------------------------------------------------------------------------------------------
	
	glEnable ( GL_COLOR_MATERIAL );
	
	glEnable ( GL_BLEND );
	
	glBlendFunc ( GL_ONE, GL_ONE );	

	//---------------------------------------------------------------------------------------------

	camera = new Camera ( Vector3D ( 0.0F, 0.0F, -18.0F ), Vector3D ( 0.0F, 0.0F, 0.0F ) );

	camera->SetFrustum ( );

	//---------------------------------------------------------------------------------------------
	
	SetupParticles ( );
	
	SetupKernels ( );
	
	BuildList ( );

	//---------------------------------------------------------------------------------------------
	
	bool running = GL_TRUE;

	int frames = 0;

	char caption [100];

	double start = glfwGetTime ( );

	//---------------------------------------------------------------------------------------------

    while ( running )
	{
        double time = glfwGetTime ( );

        if ( ( time - start ) > 1.0 || frames == 0 )
        {
            double fps = frames / ( time - start );

            sprintf_s ( caption, "N-Bodies Simulation - %.1f FPS", fps );

            glfwSetWindowTitle ( caption );

            start = time;

            frames = 0;
        }

        frames++;

		//-----------------------------------------------------------------------------------------

        glfwGetWindowSize ( &width, &height );

        height = max ( height,  1 );
		
		//-----------------------------------------------------------------------------------------

		StartKernels ( );
		
		Draw ( width, height );

		//-----------------------------------------------------------------------------------------

        glfwSwapBuffers ( );		
		
		running = !glfwGetKey ( GLFW_KEY_ESC ) && glfwGetWindowParam ( GLFW_OPENED );
	}

    glfwTerminate ( );

    return 0;
}