/*************************************************************************************************/
/*                          Fragment Shader for Updating Particle States                         */	
/*************************************************************************************************/

#extension GL_ARB_texture_rectangle : enable

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Interface

uniform sampler2DRect PositionTexture;		// Texture for storing current particle positions

uniform sampler2DRect VelocityTexture;		// Texture for storing current particle velocities

uniform vec3 SpherePosition;				// Current position of the ball

uniform float Time;							// Current time (in seconds)

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Constants

const float Step = 0.001;					// Time step

const float Attenuation = 0.8;				// Velocity attenuation at each bounce

const float Epsilon = 0.01;					// Maximum distance between bouncing bodies

const vec3 G = vec3 ( 0.0, 0.0, -9.81 );	// Particle Acceleration

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Entry Point

void main ( void )
{
	///////////////////////////////////////////////////////////////////////////////////////////////
	// Reading particle position, birth time and velocity

	vec4 position = texture2DRect ( PositionTexture, gl_TexCoord[0].xy );
	
	vec3 velocity = texture2DRect ( VelocityTexture, gl_TexCoord[0].xy ).xyz;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// If the particle already was born

	if ( Time - position.w > 0.0 )
	{
		vec3 test = position.xyz - SpherePosition;
		
		if ( abs ( length ( test ) - 2.0 ) < Epsilon )
		{	
			position.xyz += Epsilon * normalize ( test );
			
			velocity = reflect ( velocity, normalize ( test ) ) * Attenuation;
		}
		
		if ( position.z < Epsilon )
		{	
			position.z = 0.0;
			
			velocity = reflect ( velocity, vec3 ( 0.0, 0.0, 1.0 ) ) * Attenuation;
		}

		if ( position.z > 10.0 - Epsilon )
		{	
			position.z = 10.0 - Epsilon;
			
			velocity = reflect ( velocity, vec3 ( 0.0, 0.0, -1.0 ) ) * Attenuation;
		}
		
		if ( position.x < -10.0 + Epsilon )
		{	
			position.x = -10.0 + Epsilon;
			
			velocity = reflect ( velocity, vec3 ( 1.0, 0.0, 0.0 ) ) * Attenuation;
		}	
		
		if ( position.x > 10.0 - Epsilon )
		{	
			position.x = 10.0 - Epsilon;

			velocity = reflect ( velocity, vec3 ( -1.0, 0.0, 0.0 ) ) * Attenuation;
		}
		
		if ( position.y < -10.0 + Epsilon )
		{	
			position.y = -10.0 + Epsilon;
			
			velocity = reflect ( velocity, vec3 ( 0.0, 1.0, 0.0 ) ) * Attenuation;
		}	
		
		if ( position.y > 10.0 - Epsilon )
		{
			position.y = 10.0 - Epsilon;
			
			velocity = reflect ( velocity, vec3(0.0, -1.0, 0.0 )) * Attenuation;
		}	
		
		position.xyz += velocity * Step;
		
		velocity += G * Step;
	}

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Writing particle birth time, new position and velocity 
	
	gl_FragData [0] = position;
	
	gl_FragData [1] = vec4 ( velocity, 0.0 );
}