/*************************************************************************************************/
/*                             Vertex Shader for Particle Rendering                              */	
/*************************************************************************************************/

#extension GL_ARB_texture_rectangle : enable

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Interface

uniform sampler2DRect PositionTexture;		// Texture for storing current particle positions

uniform vec3 HalfSize;						// Half size of view window

uniform float Time;							// Current time (in seconds)

///////////////////////////////////////////////////////////////////////////////////////////////////
// Varying Variables for Fragment Shader

varying float Render;						// Rendering flag

varying vec4 Color;							// Base point color

varying vec3 Screen;						// Screen coords of point center

varying float PointSize;					// Point size in screen space

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Entry Point

void main ( void )
{
	vec4 position = texture2DRect ( PositionTexture, gl_MultiTexCoord0.xy );

	vec4 vertex = gl_Vertex + vec4 ( position.xyz, 0.0 );

	gl_Position = gl_ModelViewProjectionMatrix * vertex;

	gl_PointSize = max ( 1.0, 100.0 / gl_Position.w );

	Color = gl_Color;

	PointSize = gl_PointSize;

	Render = Time - position.w;

	Screen = HalfSize + HalfSize * gl_Position.xyz / gl_Position.w;
}