#include <Mouse.h>

#include <Keyboard.h>

#include <FrameBuffer.h>

using namespace Math;

using namespace Render;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ()   ()

Mouse mouse ( 0.01F );

Keyboard keyboard ( 0.05F );

Camera * camera = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int countX = 512;

int countY = 512;

///////////////////////////////////////////////////////////////////////////////////////////////////
//   ()   ,     

TextureData2D * positions = NULL;

TextureData2D * velocities = NULL;

TextureData2D * colors = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

Vector4D positionAverage ( 0.0F, 0.0F, 5.0F, 3.0F );

Vector4D positionDispersion ( 5.0F, 5.0F, 5.0F, 3.0F );

Vector3D velocityAverage ( 0.0F, 0.0F, 6.0F );

Vector3D velocityDispersion ( 5.0F, 5.0F, 2.0F );

Vector3D colorAverage ( 0.5F, 0.5F, 0.5F );

Vector3D colorDispersion ( 0.3F, 0.3F, 0.3F );

Vector3D spherePosition ( 0.0F, 0.0F, 2.0F );

///////////////////////////////////////////////////////////////////////////////////////////////////
//       

Texture2D * currentPositionTexture = NULL;

Texture2D * currentVelocityTexture = NULL;      

Texture2D * nextPositionTexture = NULL;

Texture2D * nextVelocityTexture = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ( )

ShaderManager * firstProgram = NULL;

ShaderManager * secondProgram = NULL;

ShaderManager * renderPointProgram = NULL;

ShaderManager * renderSphereProgram = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//      

FrameBuffer * firstFrameBuffer = NULL;

FrameBuffer * secondFrameBuffer = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//      [-1, +1]

float RandDouble ( void )
{
	return ( rand ( ) / ( ( float ) ( RAND_MAX ) + 1.0F ) ) * 2.0F - 1.0F;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//      

void SetupParticles ( void )
{
	positions = new TextureData2D ( countX, countY, 4 );

	velocities = new TextureData2D ( countX, countY, 3 );  

	colors = new TextureData2D ( countX, countY, 3 );
	
	for ( int i = 0; i < countX; i++ )
	{
		for ( int j = 0; j < countY; j++ )
		{
			positions->Pixel<Vector4D> ( i, j ) =
				Vector4D ( positionAverage.X + positionDispersion.X * RandDouble ( ),
				           positionAverage.Y + positionDispersion.Y * RandDouble ( ),
						   positionAverage.Z + positionDispersion.Z * RandDouble ( ),
						   positionAverage.W + positionDispersion.W * RandDouble ( ) );

			velocities->Pixel<Vector3D> ( i, j ) =
				Vector3D ( velocityAverage.X + velocityDispersion.X * RandDouble ( ),
				           velocityAverage.Y + velocityDispersion.Y * RandDouble ( ),
						   velocityAverage.Z + velocityDispersion.Z * RandDouble ( ) );

			colors->Pixel<Vector3D> ( i, j ) =
				Vector3D ( colorAverage.X + colorDispersion.X * RandDouble ( ),
				           colorAverage.Y + colorDispersion.Y * RandDouble ( ),
						   colorAverage.Z + colorDispersion.Z * RandDouble ( ) );
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//     ( )

void SetupKernels ( void )
{
	{
		currentPositionTexture = new Texture2D ( positions, 0, GL_TEXTURE_RECTANGLE_ARB );				
		currentVelocityTexture = new Texture2D ( velocities, 1, GL_TEXTURE_RECTANGLE_ARB );
		nextPositionTexture = new Texture2D ( positions, 2, GL_TEXTURE_RECTANGLE_ARB );				
		nextVelocityTexture = new Texture2D ( velocities, 3, GL_TEXTURE_RECTANGLE_ARB );
					
		currentPositionTexture->Setup ( );				
		currentVelocityTexture->Setup ( );				
		nextPositionTexture->Setup ( );		
		nextVelocityTexture->Setup ( );	;			
	}
	
	{
		firstProgram = new ShaderManager ( );
		
		if ( !firstProgram->LoadVertexShader ( "Compute.vs" ) )
		{
			cout << "ERROR! Failed to load compute vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !firstProgram->LoadFragmentShader ( "Compute.fs" ) )
		{
			cout << "ERROR! Failed to load compute fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !firstProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build compute shader program!" << endl; exit ( -1 );
		}			
		
		firstProgram->Bind ( );		
		firstProgram->SetTexture ( "PositionTexture", currentPositionTexture );		
		firstProgram->SetTexture ( "VelocityTexture", currentVelocityTexture );		
		firstProgram->Unbind ( );
	}
	
	{
		secondProgram = new ShaderManager ( );

		if ( !secondProgram->LoadVertexShader ( "Compute.vs" ) )
		{
			cout << "ERROR! Failed to load compute vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !secondProgram->LoadFragmentShader ( "Compute.fs" ) )
		{
			cout << "ERROR! Failed to load compute fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !secondProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build compute shader program!" << endl; exit ( -1 );
		}
		        
		secondProgram->Bind ( );		
		secondProgram->SetTexture ( "PositionTexture", nextPositionTexture );		
		secondProgram->SetTexture ( "VelocityTexture", nextVelocityTexture );		
		secondProgram->Unbind ( );
	}
	
	{
		renderPointProgram = new ShaderManager ( );

		if ( !renderPointProgram->LoadVertexShader ( "RenderPoint.vs" ) )
		{
			cout << "ERROR! Failed to load render vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !renderPointProgram->LoadFragmentShader ( "RenderPoint.fs" ) )
		{
			cout << "ERROR! Failed to load render fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !renderPointProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build render shader program!" << endl; exit ( -1 );
		}
		
		renderPointProgram->Bind ( );		
		renderPointProgram->SetTexture ( "PositionTexture", currentPositionTexture );		
		renderPointProgram->Unbind ( );
	}
	
	{
		renderSphereProgram = new ShaderManager ( );

		if ( !renderSphereProgram->LoadVertexShader ( "RenderSphere.vs" ) )
		{
			cout << "ERROR! Failed to load render vertex shader!" << endl; exit ( -1 );
		}
		
		if ( !renderSphereProgram->LoadFragmentShader ( "RenderSphere.fs" ) )
		{
			cout << "ERROR! Failed to load render fragment shader!" << endl; exit ( -1 );
		}
		
		if ( !renderSphereProgram->BuildProgram ( ) )
		{
			cout << "ERROR! Failed to build render shader program!" << endl; exit ( -1 );
		}
	}
	
	{
		firstFrameBuffer = new FrameBuffer ( );		
		firstFrameBuffer->ColorBuffers.push_back ( nextPositionTexture );
		firstFrameBuffer->ColorBuffers.push_back ( nextVelocityTexture );
		firstFrameBuffer->Setup ( );
	}
	
	{
		secondFrameBuffer = new FrameBuffer ( );		
		secondFrameBuffer->ColorBuffers.push_back ( currentPositionTexture );		
		secondFrameBuffer->ColorBuffers.push_back ( currentVelocityTexture );
		secondFrameBuffer->Setup ( );
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//       ( = )

void GenerateStream ( int width, int height )
{
	{
		glViewport ( 0, 0, width, height );    
				
		glMatrixMode ( GL_PROJECTION );
	            
		glLoadIdentity ( ); 
		
		glOrtho ( 0.0F, 1.0F, 0.0F, 1.0F, -1.0F, 1.0F );
		
		glMatrixMode ( GL_MODELVIEW );
		
		glLoadIdentity ( );				
	}

	{
		glClear ( GL_COLOR_BUFFER_BIT );
		
		glBegin ( GL_QUADS );
		{
			glTexCoord2f (   0.0F,   0.0F );  glVertex2f ( 0.0F, 0.0F );
			glTexCoord2f (   0.0F, countY );  glVertex2f ( 0.0F, 1.0F );
			glTexCoord2f ( countX, countY );  glVertex2f ( 1.0F, 1.0F );
			glTexCoord2f ( countX,   0.0F );  glVertex2f ( 1.0F, 0.0F );
		}
		glEnd ( ); 
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

bool step = true;

void StartKernels ( float time )
{
	if ( step = !step )
	{
		firstFrameBuffer->Bind ( );

		firstProgram->Bind ( );

		firstProgram->SetUniformFloat ( "Time", time );

		firstProgram->SetUniformVector ( "SpherePosition", spherePosition );

		GenerateStream ( countX, countY );

		firstFrameBuffer->Unbind ( );
	}
	else
	{
		secondFrameBuffer->Bind ( );

		secondProgram->Bind ( );

		secondProgram->SetUniformFloat ( "Time", time );

		secondProgram->SetUniformVector ( "SpherePosition", spherePosition );

		GenerateStream ( countX, countY );

		secondFrameBuffer->Unbind ( );
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//       

void BuildList ( void )
{
	glEnable ( GL_VERTEX_PROGRAM_POINT_SIZE );
	
	glEnable ( GL_POINT_SMOOTH );

	glNewList ( 1, GL_COMPILE );

	glBegin ( GL_POINTS );

	for ( int i = 0; i < countX; i++ )
	{
		for ( int j = 0; j < countY; j++ )
		{
			glColor3fv ( colors->Pixel<Vector3D> ( i, j ) );

			glTexCoord2f ( i, j );
			
			glVertex3f ( 0.0F, 0.0F, 0.0F );
		}
	}

	glEnd ( );

	glEndList ( );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  

void Draw ( int width, int height, float time )
{
	glEnable ( GL_DEPTH_TEST );	
	
	{
		mouse.Apply ( camera );

		keyboard.Apply ( camera );

		camera->SetViewport ( width, height );
	            	
		camera->Setup ( );
	}

	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	{
		renderPointProgram->Bind ( );

		renderPointProgram->SetUniformFloat (
			"Time", time );

		renderPointProgram->SetUniformVector (
			"HalfSize", Vector3D ( width, height, 1.0F ) / 2.0F );

		glCallList ( 1 );

		renderPointProgram->Unbind ( );
	}

	{
		glColor3f ( 0.8F, 0.8F, 0.8F );

		glBegin ( GL_LINE_LOOP );		

			glVertex3f ( -10.0F, -10.0F, 0.0F );
			glVertex3f (  10.0F, -10.0F, 0.0F );
			glVertex3f (  10.0F,  10.0F, 0.0F );
			glVertex3f ( -10.0F,  10.0F, 0.0F );		

		glEnd ( );

		glBegin ( GL_LINE_LOOP );	

			glVertex3f ( -10.0F, -10.0F, 10.0F );
			glVertex3f (  10.0F, -10.0F, 10.0F );
			glVertex3f (  10.0F,  10.0F, 10.0F );
			glVertex3f ( -10.0F,  10.0F, 10.0F );		

		glEnd ( );

		glBegin ( GL_LINES );			

			glVertex3f ( -10.0F, -10.0F,  0.0F );
			glVertex3f ( -10.0F, -10.0F, 10.0F );
			glVertex3f (  10.0F, -10.0F,  0.0F );
			glVertex3f (  10.0F, -10.0F, 10.0F );
			glVertex3f (  10.0F,  10.0F,  0.0F );
			glVertex3f (  10.0F,  10.0F, 10.0F );
			glVertex3f ( -10.0F,  10.0F,  0.0F );
			glVertex3f ( -10.0F,  10.0F, 10.0F );	

		glEnd ( );
	}

	{
		GLUquadric * quad = gluNewQuadric ( );

		gluQuadricTexture ( quad, true );  

		glTranslatef ( spherePosition.X,
			           spherePosition.Y,
					   spherePosition.Z );

		renderSphereProgram->Bind ( );

		gluSphere ( quad, 2.0F, 30, 30 );

		renderSphereProgram->Unbind ( );

		gluDeleteQuadric ( quad );
	}

	glDisable ( GL_DEPTH_TEST );	
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

void MouseMove ( int x, int y )
{
	mouse.MouseMove ( x, y );
}

void MouseButton ( int button, int state )
{
	mouse.StateChange ( button, state );
}

void KeyButton ( int key, int state )
{
	keyboard.StateChange ( key, state );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int main ( void )
{
	glfwInit ( );

	//---------------------------------------------------------------------------------------------

	cout << "Do you want to run program in fullscreen mode? [Y/N]" << endl;

	int choice = getchar ( );

	int width = 750, height = 750, mode = GLFW_WINDOW;

	//---------------------------------------------------------------------------------------------

	if ( choice == 'Y' || choice == 'y' )
	{
		GLFWvidmode vidmode;

		glfwGetDesktopMode ( &vidmode );

		width = vidmode.Width;
		
		height = vidmode.Height;

		mode = GLFW_FULLSCREEN;
	}

	//---------------------------------------------------------------------------------------------

    if( !glfwOpenWindow ( width, height, 0, 0, 0, 0, 16, 0, mode ) )
    {
        glfwTerminate ( ); exit ( 0 );
	}

	//---------------------------------------------------------------------------------------------

    glfwSwapInterval ( 0 );

	glfwSetMousePosCallback ( MouseMove );

	glfwSetMouseButtonCallback ( MouseButton );

	glfwSetKeyCallback ( KeyButton );

	//---------------------------------------------------------------------------------------------

	camera = new Camera ( Vector3D ( -18.0F, -18.0F, 8.0F ),
		                  Vector3D ( 1.3F, 2.37F, 0.19F ) );

	camera->SetFrustum ( );

	//---------------------------------------------------------------------------------------------
	
	SetupParticles ( );
	
	SetupKernels ( );

	BuildList ( );

	//---------------------------------------------------------------------------------------------
	
	bool running = GL_TRUE;

	int frames = 0;

	char caption [100];

	double start = glfwGetTime ( );

	//---------------------------------------------------------------------------------------------

    while ( running )
	{
        double time = glfwGetTime ( );

        if ( ( time - start ) > 1.0 || frames == 0 )
        {
            double fps = frames / ( time - start );

            sprintf_s ( caption, "Simple Particle System - %.1f FPS", fps );

            glfwSetWindowTitle ( caption );

            start = time;

            frames = 0;
        }

        frames++;

		//-----------------------------------------------------------------------------------------

        glfwGetWindowSize ( &width, &height );

        height = max ( height,  1 );
		
		//-----------------------------------------------------------------------------------------

		spherePosition.X = 5.0F * sinf ( 6.0F * time );
		spherePosition.Y = 5.0F * cosf ( 6.0F * time );

		StartKernels ( time );
		
		Draw ( width, height, time );

		//-----------------------------------------------------------------------------------------

        glfwSwapBuffers ( );		
		
		running = !glfwGetKey ( GLFW_KEY_ESC ) && glfwGetWindowParam ( GLFW_OPENED );
	}

    glfwTerminate ( ); exit ( 0 );
}