#include <ShaderManager.h>

#include <Mouse.h>

#include <Keyboard.h>

#include <Texture2D.h>

#include <Texture3D.h>

#include <Texture3D.h>

#include <FrameBuffer.h>

using namespace Render;

using namespace Math;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ()   ()

Mouse mouse ( 0.01F );

Keyboard keyboard ( 0.05F );

Camera * camera = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//   

int mapWidth = 80;

int mapHeight = 80;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

ShaderManager * rayTracingManager = NULL;

ShaderManager * photonMappingManager = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//        ( )

FrameBuffer * photonMappingFrameBuffer = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//       

Texture2D * photonTexture = NULL;

Texture2D * bottomTexture = NULL;

Texture2D * bottomBumpTexture = NULL;

Texture2D * wallTexture = NULL;

Texture2D * wallBumpTexture = NULL;

Texture2D * ceilingTexture = NULL;

Texture2D * ceilingBumpTexture = NULL;

///////////////////////////////////////////////////////////////////////////////////////////////////
//    a  b

int Compare ( const void * a, const void * b )
{
	Vector3D * left = ( Vector3D * ) a;
	
	Vector3D * right = ( Vector3D * ) b;

	if ( left->X > right->X ) return 1;
	
	else if ( left->X == right->X && left->Y > right->Y ) return 1;
	
	else if ( left->X == right->X && left->Y == right->Y && left->Z > right->Z ) return 1;
	
	else if ( left->X == right->X && left->Y == right->Y && left->Z == right->Z ) return 0;
	
	return -1;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//        

void MouseMove ( int x, int y )
{
	mouse.MouseMove ( x, y );
}
		
void MouseButton ( int button, int state )
{
	mouse.StateChange ( button, state );
}

void KeyButton ( int key, int state )
{
	keyboard.StateChange ( key, state );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//    

int main ( void )
{
	glfwInit ( );

	//---------------------------------------------------------------------------------------------

	cout << "Do you want to run program in fullscreen mode? [Y/N]" << endl;

	int choice = getchar ( );

	int width = 750, height = 750, mode = GLFW_WINDOW;

	//---------------------------------------------------------------------------------------------

	if ( choice == 'Y' || choice == 'y' )
	{
		GLFWvidmode vidmode;
		glfwGetDesktopMode ( &vidmode );

		width = vidmode.Width;		
		height = vidmode.Height;

		mode = GLFW_FULLSCREEN;
	}

	//---------------------------------------------------------------------------------------------

    if( !glfwOpenWindow ( width, height, 0, 0, 0, 0, 0, 0, mode ) )
    {
        glfwTerminate ( ); exit ( 0 );
	}

	//---------------------------------------------------------------------------------------------

    glfwSwapInterval ( 0 );
	glfwSetMousePosCallback ( MouseMove );
	glfwSetMouseButtonCallback ( MouseButton );
	glfwSetKeyCallback ( KeyButton );

	//---------------------------------------------------------------------------------------------

	photonTexture = new Texture2D (
		new TextureData2D ( mapWidth, mapHeight, 3 ), 0 , GL_TEXTURE_RECTANGLE_ARB );
	photonTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	bottomTexture = new Texture2D (
		TextureData2D :: FromTGA ( "Bottom.tga" ) , 1, GL_TEXTURE_2D );
	bottomTexture->FilterMode = FilterMode::Linear;
	bottomTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	bottomBumpTexture = new Texture2D (
		TextureData2D :: FromTGA ( "BottomBump.tga" ) , 2, GL_TEXTURE_2D );
	bottomBumpTexture->FilterMode = FilterMode::Linear;
	bottomBumpTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	wallTexture = new Texture2D (
		TextureData2D :: FromTGA ( "Wall2.tga" ) , 3, GL_TEXTURE_2D );
	wallTexture->FilterMode = FilterMode::Linear;
	wallTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	wallBumpTexture = new Texture2D (
		TextureData2D :: FromTGA ( "WallBump2.tga" ) , 4, GL_TEXTURE_2D );
	wallBumpTexture->FilterMode = FilterMode::Linear;
	wallBumpTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	ceilingTexture = new Texture2D (
		TextureData2D :: FromTGA ( "Ceiling1.tga" ) , 5, GL_TEXTURE_2D );
	ceilingTexture->FilterMode = FilterMode::Linear;
	ceilingTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	ceilingBumpTexture = new Texture2D (
		TextureData2D :: FromTGA ( "CeilingBump1.tga" ) , 6, GL_TEXTURE_2D );
	ceilingBumpTexture->FilterMode = FilterMode::Linear;
	ceilingBumpTexture->Setup ( );

	//---------------------------------------------------------------------------------------------

	photonMappingFrameBuffer = new FrameBuffer ( );
	photonMappingFrameBuffer->ColorBuffers.push_back ( photonTexture );
	photonMappingFrameBuffer->Setup ( );

	//---------------------------------------------------------------------------------------------

	rayTracingManager = new ShaderManager ( );
	rayTracingManager->LoadVertexShader ( "Render.vs" );
	rayTracingManager->LoadFragmentShader ( "Render.fs" );
	rayTracingManager->BuildProgram ( );

	//---------------------------------------------------------------------------------------------
	
	photonMappingManager = new ShaderManager ( );
	photonMappingManager->LoadVertexShader ( "Render.vs" );
	photonMappingManager->LoadFragmentShader ( "Render.fs", "#define PHOTON_MAP" );
	photonMappingManager->BuildProgram ( );

	//---------------------------------------------------------------------------------------------

	rayTracingManager->Bind ( );
	rayTracingManager->SetUniformVector ( "BoxMinimum", Vector3D ( -5.0F, -5.0F, -5.0F ) );
	rayTracingManager->SetUniformVector ( "BoxMaximum", Vector3D ( 5.0F, 5.0F, 5.0F ) );	
	rayTracingManager->SetUniformVector ( "Sphere.Center", Vector3D ( -3.0F, -3.0F, -3.0F ) );
	rayTracingManager->SetUniformFloat ( "Sphere.Radius", 2.0F );
	rayTracingManager->SetUniformVector ( "Light.Position", Vector3D ( 0.0F, 0.0F, 0.0F ) );	
	rayTracingManager->SetTexture ( "PhotonTexture", photonTexture );
	rayTracingManager->SetTexture ( "BottomTexture", bottomTexture );
	rayTracingManager->SetTexture ( "BottomBumpTexture", bottomBumpTexture );
	rayTracingManager->SetTexture ( "WallTexture", wallTexture );
	rayTracingManager->SetTexture ( "WallBumpTexture", wallBumpTexture );
	rayTracingManager->SetTexture ( "CeilingTexture", ceilingTexture );
	rayTracingManager->SetTexture ( "CeilingBumpTexture", ceilingBumpTexture );	
	rayTracingManager->SetUniformFloat ( "Delta", 0.2F  );
	rayTracingManager->SetUniformFloat ( "InverseDelta", 1.0F / 0.2F );
	rayTracingManager->SetUniformVector ( "PhotonMapSize", Vector2D ( mapWidth, mapHeight ) );
	rayTracingManager->SetUniformFloat ( "PhotonIntensity", 300.0F / ( mapWidth * mapHeight ) );
	rayTracingManager->Unbind ( );

	//---------------------------------------------------------------------------------------------

	photonMappingManager->Bind ( );
	photonMappingManager->SetUniformVector ( "BoxMinimum", Vector3D ( -5.0F, -5.0F, -5.0F ) );
	photonMappingManager->SetUniformVector ( "BoxMaximum", Vector3D ( 5.0F, 5.0F, 5.0F ) );
	photonMappingManager->SetUniformVector ( "Sphere.Center", Vector3D ( -3.0F, -3.0F, -3.0F ) );
	photonMappingManager->SetUniformFloat ( "Sphere.Radius", 2.0F );
	photonMappingManager->SetUniformVector ( "Light.Position", Vector3D ( 0.0F, 0.0F, 0.0F ) );
	photonMappingManager->SetUniformVector ( "Light.Radius", Vector2D ( 0.5F * 10, 0.5F * 10 ) );
	photonMappingManager->SetUniformFloat ( "Light.Distance", 0.5F * 10 );
	photonMappingManager->Unbind ( );

	//---------------------------------------------------------------------------------------------
	
	glMatrixMode ( GL_PROJECTION );
	
	glLoadIdentity ( );
	
	glOrtho ( -1.0F, 1.0F, -1.0F, 1.0F, -1.0F, 1.0F  );
	
	glMatrixMode ( GL_MODELVIEW );
	
	glLoadIdentity ( );	

	//---------------------------------------------------------------------------------------------

	camera = new Camera ( Vector3D ( 0.0F, 0.0F, -18.0F ),
		                  Vector3D ( 0.0F, 0.0F, 0.0F ) );

	camera->SetFrustum ( );

	//---------------------------------------------------------------------------------------------
	
	bool running = GL_TRUE;

	int frames = 0;

	char caption [100];

	double start = glfwGetTime ( );

	//---------------------------------------------------------------------------------------------

    while ( running )
    {
        double time = glfwGetTime ( );

        if ( ( time - start ) > 1.0 || frames == 0 )
        {
            double fps = frames / ( time - start );

            sprintf_s ( caption, "Two-Pass Ray Tracing - %.1f FPS", fps );

            glfwSetWindowTitle ( caption );

            start = time;

            frames = 0;
        }

        frames++;

		//-----------------------------------------------------------------------------------------

        glfwGetWindowSize ( &width, &height );

        height = max ( height, 1 );

		camera->SetViewport ( width, height );
		
		//-----------------------------------------------------------------------------------------

		mouse.Apply ( camera );

		keyboard.Apply ( camera );

		//-----------------------------------------------------------------------------------------

		photonMappingFrameBuffer->Bind ( );

		photonMappingManager->Bind ( );

		photonMappingManager->SetUniformVector ( "Sphere.Center",
			Vector3D ( ( 2.0F + sinf ( time / 5.0F ) ) * sinf ( time ),
			           -2.0F + sinf ( time / 5.0F ),
					   ( 2.0F + sinf ( time / 5.0F ) )  * cosf ( time ) ) );
		
		glViewport ( 0, 0, mapWidth, mapHeight );

		glClear ( GL_COLOR_BUFFER_BIT );

		glBegin ( GL_QUADS );

			glVertex2f ( -1.0F, -1.0F );   glVertex2f ( -1.0F,  1.0F );
			glVertex2f (  1.0F,  1.0F );   glVertex2f (  1.0F, -1.0F );

		glEnd ( );

		photonMappingManager->Unbind ( );

		photonMappingFrameBuffer->FetchOutcome ( );

		photonMappingFrameBuffer->Unbind ( );

		//-----------------------------------------------------------------------------------------

		qsort ( *photonTexture->Data,
				photonTexture->Data->GetWidth ( ) * photonTexture->Data->GetHeight ( ),
				sizeof ( Vector3D ),
				Compare );

		photonTexture->Update ( );

		//-----------------------------------------------------------------------------------------

		rayTracingManager->Bind ( );

		rayTracingManager->SetUniformVector ( "Sphere.Center",
			Vector3D ( ( 2.0F + sinf ( time / 5.0F ) ) * sinf ( time ),
			           -2.0F + sinf ( time / 5.0F ),
					   ( 2.0F + sinf ( time / 5.0F ) ) * cosf ( time ) ) );

		camera->SetShaderData ( rayTracingManager );

		glViewport ( 0, 0, width, height );

		glClear ( GL_COLOR_BUFFER_BIT );

		glBegin ( GL_QUADS );

			glVertex2f ( -1.0F, -1.0F );   glVertex2f ( -1.0F,  1.0F );
			glVertex2f (  1.0F,  1.0F );   glVertex2f (  1.0F, -1.0F );

		glEnd ( );

		rayTracingManager->Unbind ( );

        glfwSwapBuffers ( );

		//-----------------------------------------------------------------------------------------
		
		running = !glfwGetKey ( GLFW_KEY_ESC ) && glfwGetWindowParam ( GLFW_OPENED );
	}

    glfwTerminate ( ); exit ( 0 );
}